from typing import Dict

from drkns.stepexecutionstatus.StepExecutionStatus import StepExecutionStatus


class _Context:

    def __init__(self):
        self.cached_execution_status: Dict[str, StepExecutionStatus] = {}


context: _Context = _Context()



from typing import List, Tuple

from drkns.stepexecutionstatus.StepExecutionStatus import StepExecutionStatus
from drkns.stepexecutionstatus.get_status_message_and_output_from_status import\
    get_status_message_and_output_from_status


def get_successful_flag_and_combined_output(
            execution_history: List[StepExecutionStatus],
            summary: bool = False,
            limit_output: bool = False
            ) -> Tuple[bool, List[str]]:
    outputs = []
    statuses = []
    successful = True
    for status in execution_history:
        message, output = get_status_message_and_output_from_status(status)

        statuses.append(message)

        if status.ignored or not status.successful:
            successful = False

        active_output = not status.ignored and not status.restored
        append_output = not limit_output or active_output
        if output is not None and append_output:
            outputs.append(output)

    if summary:
        return successful, statuses

    combined_output = []
    if len(outputs) > 0:
        combined_output += outputs + ['\n']

    combined_output += statuses

    return successful, combined_output



from typing import Tuple


from drkns.generation.templateloading.get_generation_template_path \
    import get_generation_template_path
import drkns.generation.pattern_util as pattern_util
from drkns.generation.GenerationTemplate import GenerationTemplate
from drkns.exception import UnableToParseGenerationTemplateBlock
from drkns.generation._tags import unit_template_tag_prefix, group_template_tag_prefix
from drkns.generation.templateloading.check_generation_template \
    import check_generation_template


def get_generation_template(from_path: str) -> GenerationTemplate:
    with open(get_generation_template_path(from_path)) as handle:
        raw_template = handle.read()
        generation_template = _parse(from_path, raw_template)
        check_generation_template(generation_template)
        return generation_template


def _parse(from_path, raw_template) -> GenerationTemplate:
    unit_template, raw_template =\
        _extract_from_tag_prefix(unit_template_tag_prefix, raw_template)
    group_template, template = \
        _extract_from_tag_prefix(group_template_tag_prefix, raw_template)
    generation_template = GenerationTemplate(
        from_path, template, group_template, unit_template)
    return generation_template


def _extract_from_tag_prefix(tag_prefix: str, content: str) -> Tuple[str, str]:
    block_re = pattern_util.get_block_re_from_tag_prefix(tag_prefix)
    match = block_re.search(content)
    if match is None:
        error_message = 'No block found from prefix ' + tag_prefix
        raise UnableToParseGenerationTemplateBlock(error_message)
    start_index = match.start()
    end_index = match.end()
    _, block, _ = match.groups()
    stripped_content = content[:start_index] + content[end_index:]
    return block, stripped_content





class GenerationTemplate:

    def __init__(
            self,
            source_path: str,
            template: str,
            group_template: str,
            unit_template: str
    ):
        self.source_path: str = source_path
        self.template: str = template
        self.group_template: str = group_template
        self.unit_template: str = unit_template



import yaml
import os
import re
from typing import Optional, List

from drkns.exception import MalformedIgnorePatternException
from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.configunit import config_directory

_ignored_by_default = [
    '.git', '.git/', '.drknspersistence/', '*.pyc', '__pycache__/'
]


def load(
        root_path: str,
        inherited_ignored: Optional[List[str]] = None,
        original_root_path: Optional[str] = None
        ) -> ConfigUnit:
    root_path = os.path.abspath(root_path)
    if original_root_path is None:
        original_root_path = os.path.abspath(root_path)

    if not os.path.isdir(original_root_path):
        original_root_path = os.path.dirname(original_root_path)

    if inherited_ignored is None:
        inherited_ignored = []

    unit_name = _get_unit_name(root_path, original_root_path)

    if os.path.isdir(root_path):
        root_path = os.path.join(root_path, 'drkns.yml')

    with open(root_path) as handle:
        data = yaml.safe_load(handle)

    base_dir = os.path.dirname(root_path)
    if 'directory' in data:
        data['directory'] = \
            os.path.abspath(os.path.join(base_dir, data['directory']))
    else:
        data['directory'] = base_dir

    config_unit_ignored = _get_ignored(root_path)
    parsed_ignored = inherited_ignored + config_unit_ignored

    raw_dependencies = data.get('dependencies', [])

    if isinstance(raw_dependencies, dict):
        error_message = 'dependencies must be an array of path now, ' + \
                        'error in: ' + root_path
        raise Exception(error_message)

    parsed_dependencies = []
    for relative_path in raw_dependencies:
        target_path = os.path.abspath(os.path.join(base_dir, relative_path))
        if not (target_path in config_directory):
            dependency = load(target_path, parsed_ignored, original_root_path)
        else:
            dependency = config_directory[target_path]

        parsed_dependencies.append(dependency)

    data['dependencies'] = parsed_dependencies

    ignored = parsed_ignored + _ignored_by_default
    config_unit = ConfigUnit(unit_name, data, ignored)

    config_directory[root_path] = config_unit

    return config_unit


def _get_unit_name(config_path: str, original_root_path: str) -> str:
    if not os.path.isdir(config_path):
        config_path = os.path.dirname(config_path)

    if config_path == original_root_path:
        return 'main'

    unit_name = \
        os.path.relpath(config_path, original_root_path)

    unit_name = unit_name.replace('/', '_')
    unit_name = unit_name.replace('.', '_')

    if unit_name == '_':
        import sys
        sys.exit()

    return unit_name


def _get_ignored(path: str) -> List[str]:
    dir_name = path
    if not os.path.isdir(dir_name):
        dir_name = os.path.dirname(dir_name)

    drkns_ignore_path = os.path.join(dir_name, '.drknsignore')
    if not os.path.exists(drkns_ignore_path):
        return []

    handle = open(drkns_ignore_path)
    ignored_raw = handle.read()
    ignored_raw = ignored_raw.replace('\r', ' ')
    ignored_raw = ignored_raw.replace('\n', ' ')
    ignored_raw = ignored_raw.replace(',', ' ')
    separator = u' '
    ignored_raw = re.sub(r'\s\s+', separator, ignored_raw)

    ignored = ignored_raw.split(separator)
    _check_ignored(ignored)

    return ignored


def _check_ignored(ignored: List[str]):
    for ignored_item in ignored:
        slash_index = ignored_item.find('/')
        if slash_index == -1:
            continue
        if slash_index == len(ignored_item) -1:
            continue

        error_message = '"/" is only allowed at the end of a pattern, found "' \
                        + ignored_item + '"'
        MalformedIgnorePatternException(
            error_message)


import pickle

from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.stepexecutionstatus.StepExecutionStatus import StepExecutionStatus
from drkns.context.Context import context
from drkns.context.get_unit_step_path import get_unit_step_path


def store_past_execution_status(
        config_unit: ConfigUnit,
        step_execution_status: StepExecutionStatus):
    persistence_path = get_unit_step_path(
        config_unit, step_execution_status.step_name)
    with open(persistence_path, 'wb') as persistence_file:
        pickle.dump(step_execution_status, persistence_file)

    context.cached_execution_status[persistence_path] = step_execution_status



import os
import re

from drkns.exception import MissingGenerationTemplateDirectoryException, \
    MissingGenerationTemplateException, MultipleGenerationTemplateException


_template_directory = '.drknsgeneration'
_template_file_re = re.compile(r'^.*\.template\..*$')


def get_generation_template_path(from_path: str) -> str:
    template_directory = os.path.join(from_path, _template_directory)
    if not os.path.exists(template_directory):
        error_message = template_directory + ' directory can not be found'
        raise MissingGenerationTemplateDirectoryException(error_message)

    contained_files = os.listdir(template_directory)

    matched_files = []
    for contained_file in contained_files:
        pattern_match = \
            _template_file_re.match(contained_file)
        if pattern_match is None:
            continue

        matched_files.append(contained_file)

    if len(matched_files) == 0:
        error_message = 'No template found in ' + template_directory
        raise MissingGenerationTemplateException(error_message)

    if len(matched_files) > 1:
        error_message = 'Multiple template found in ' + \
                        template_directory + ': ' + ', '.join(matched_files)
        MultipleGenerationTemplateException(error_message)

    return os.path.join(template_directory, matched_files.pop())


from typing import List

import re

from drkns.exception import UnableToFindPattern


def format_list_in_template(
        tag: str,
        items: List[str],
        template: str,
        line_level: bool
) -> str:
    list_re = get_list_re_from_tag(tag, line_level)
    list_match = list_re.search(template)
    if list_match is None:
        raise UnableToFindPattern(list_re.pattern)

    separator = list_match.groupdict()['separator']
    joined_items = separator.join(items)
    start_index = list_match.start()
    end_index = list_match.end()
    return template[:start_index] + joined_items + template[end_index:]


def format_list_optional_in_template(
        tag: str,
        items: List,
        template: str
) -> str:
    list_optional_re = get_list_optional_re_from_tag(tag)
    list_optional_match = list_optional_re.search(template)
    if list_optional_match is None:
        raise UnableToFindPattern(list_optional_re.pattern)

    conditional_value = ''
    if len(items) > 0:
        conditional_value = list_optional_match.groupdict()['conditional_value']

    start_index = list_optional_match.start()
    end_index = list_optional_match.end()
    return template[:start_index] + conditional_value + template[end_index:]


def get_list_re_from_tag(tag: str, line_level: bool) -> re.Pattern:
    pattern = get_list_replacement_pattern_from_tag(tag, line_level)
    return re.compile(pattern)


def get_list_optional_re_from_tag(tag: str) -> re.Pattern:
    pattern = r'%' + tag + r'\?\{(?P<conditional_value>.*)\}%'
    return re.compile(pattern)


def get_list_replacement_pattern_from_tag(tag: str, line_level: bool) -> str:
    pattern = '%' + tag + r'\{(?P<separator>.*)\}%'
    if not line_level:
        return pattern
    return _get_line_level_pattern(pattern)


def get_block_re_from_tag_prefix(tag_prefix: str) -> re.Pattern:
    begin_line_pattern = get_line_pattern_from_tag(tag_prefix + 'BEGIN')
    end_line_pattern = get_line_pattern_from_tag(tag_prefix + 'END')
    block_pattern = \
        '(?P<begin_line>' + begin_line_pattern + ')' +\
        '(?P<block>(?:.|\n)*)' +\
        '(?P<end_line>' + end_line_pattern + ')'
    return re.compile(block_pattern)


def get_line_pattern_from_tag(tag: str) -> str:
    return _get_line_level_pattern('%' + tag + '%')


def get_pattern_from_tag(tag: str) -> str:
    return '%' + tag + '%'


def _get_line_level_pattern(pattern: str) -> str:
    return '(?:^|\n).*' + pattern + '.*[\n$]'



from typing import Tuple, Optional

from drkns.stepexecutionstatus.StepExecutionStatus import StepExecutionStatus
from drkns.util import BColors


def get_status_message_and_output_from_status(status: StepExecutionStatus)\
        -> Tuple[str, Optional[str]]:
    """

    :param status:
    :return: status_message and output (can be None)
    """
    output = None
    message = status.name() + ': ' + BColors.BOLD
    if status.ignored:
        message += BColors.WARNING + 'Ignored' + BColors.ENDC
    elif not status.successful:
        message += BColors.FAIL + 'Error' + BColors.ENDC
    else:
        message += BColors.OKBLUE + 'OK' + BColors.ENDC

    if status.restored:
        message += ' (restored)'

    if not status.successful:
        output = 'Output for ' + status.name()

        if status.restored:
            output += ' (restored)'

        output += ': \n' + status.output + '\n--\n\n'

    return message, output



from drkns.configunit.ConfigUnit import ConfigUnit
import drkns.step.step_type as step_type


def get_step_type(config_unit: ConfigUnit, step_name: str) -> str:
    for possible_step_type in step_type.types:
        step_names = config_unit.get_steps(possible_step_type)
        if step_name in step_names:
            return possible_step_type

    raise Exception('Unable to resolve type for step ' + step_name +
                    ' in ConfigUnit ' + config_unit.name)




class UnknownDependencyException(Exception):
    pass


class CircularDependencyException(Exception):
    pass


class UnknownStepException(Exception):
    pass


class UnknownCommandException(Exception):
    pass


class UnknownCommandFlagException(Exception):
    pass


class MissingCommandException(Exception):
    pass


class MissingForgetTargetException(Exception):
    pass


class MissingSyncDirectionException(Exception):
    pass


class MissingS3PathException(Exception):
    pass


class MalformedIgnorePatternException(Exception):
    pass


class MissingGenerationTemplateDirectoryException(Exception):
    pass


class MissingGenerationTemplateException(Exception):
    pass


class MultipleGenerationTemplateException(Exception):
    pass


class UnableToParseGenerationTemplateBlock(Exception):
    pass


class MissingGenerationTemplateTag(Exception):
    pass


class UnableToFindPattern(Exception):
    pass


class DependenciesNotAvailable(Exception):
    pass



import copy
from typing import Tuple, List, Dict, Optional


from drkns.util import sh
from drkns.configunit import ConfigUnit
from drkns.stepexecutionstatus.StepExecutionStatus import StepExecutionStatus

from drkns.step.get_step_type import get_step_type
import drkns.step.step_type
from drkns.context.get_past_execution_status import \
    get_past_execution_status
from drkns.context.store_past_execution_status import \
    store_past_execution_status


def run_plan(plan: List[Tuple[ConfigUnit, str, str]]) \
        -> List[StepExecutionStatus]:
    return _PlanRunner(plan).run()


class _PlanRunner:

    def __init__(self, plan: List[Tuple[ConfigUnit, str, str]]):
        self._plan: List[Tuple[ConfigUnit, str, str]] = plan
        self._status_history: List[StepExecutionStatus] = []
        self._failed_dependency_check_statuses: Dict[str, StepExecutionStatus] = {}
        self._failed_dependency_build_statuses: Dict[str, StepExecutionStatus] = {}
        self._processing_config_unit: Optional[ConfigUnit] = None

    def run(self) -> List[StepExecutionStatus]:
        self._status_history = []
        for config_unit, step_name, prefixed_step_name in self._plan:
            self._clean_up_pending_processes_if_needed(config_unit)
            step_execution_status = \
                self._get_execution_status_from_past_execution(
                    config_unit, step_name)

            if step_execution_status is None:
                step_execution_status = self._run(config_unit, step_name)

            self._handle_execution_status(config_unit, step_execution_status)

        self._clean_up_pending_processes_if_needed()

        return self._status_history

    # BEGIN cleanup: Remove pending processes than can be spawned by steps when
    # they are not needed anymore

    def _clean_up_pending_processes_if_needed(
        self,
        current_config_unit: ConfigUnit = None
    ):
        need_process_cleanup = self._need_to_cleanup_pending_processes(
            current_config_unit)

        to_cleanup_config_unit = self._processing_config_unit
        self._processing_config_unit = current_config_unit
        if (not need_process_cleanup) or (to_cleanup_config_unit is None):
            # `to_cleanup_config_unit` is None is here to silence mypy, it's
            # already checked in need_process_cleanup
            return

        while len(to_cleanup_config_unit.pending_subprocesses) != 0:
            subprocess = to_cleanup_config_unit.pending_subprocesses.pop(0)
            subprocess.terminate()

    def _need_to_cleanup_pending_processes(
            self,
            current_config_unit: Optional[ConfigUnit]
            ) -> bool:
        if self._processing_config_unit is None:
            return False

        if current_config_unit is None:
            return True

        new_config_unit = current_config_unit.name != \
            self._processing_config_unit.name
        return new_config_unit

    # END cleanup

    # noinspection PyMethodMayBeStatic
    def _run(self, config_unit: ConfigUnit, step_name: str) \
            -> StepExecutionStatus:
        """
        Actually calls the underlying command

        :param config_unit:
        :param step_name:
        :return:
        """
        step_type = get_step_type(config_unit, step_name)
        step = config_unit.get_steps(step_type)[step_name]

        # Run command
        command = '(cd "' + config_unit.directory + '"; \n' + step.command +\
                  '\n)'
        return_code, output, subprocess = sh(command, detached=step.background)
        successful = return_code == 0

        if subprocess.poll() is None:
            config_unit.pending_subprocesses.append(subprocess)

        step_execution_status = \
            StepExecutionStatus(config_unit.name, step_name,
                                successful=successful, output=output,
                                step_type=step_type)

        return step_execution_status

    def _handle_execution_status(
            self,
            config_unit: ConfigUnit,
            step_execution_status: StepExecutionStatus):

        step_type = step_execution_status.step_type
        is_check = step_type == drkns.step.step_type.CHECK
        successful = step_execution_status.successful
        if (not successful) and is_check:
            self._failed_dependency_check_statuses[config_unit.name] = \
                step_execution_status

        is_build = step_type == drkns.step.step_type.BUILD
        if (not successful) and is_build:
            self._failed_dependency_build_statuses[config_unit.name] = \
                step_execution_status

        if not step_execution_status.restored:
            store_past_execution_status(config_unit, step_execution_status)

        self._status_history.append(step_execution_status)

    # BEGIN past status retrieval:
    #   - Either from a previous execution without dependent code change
    #   - Either from a step that failed on which this step depends

    def _get_execution_status_from_past_execution(
            self,
            config_unit: ConfigUnit,
            step_name: str
            ) -> Optional[StepExecutionStatus]:
        """
        Will resolve a persisted execution status or a failed one from
        dependency execution
        :param config_unit:
        :param step_name:
        :return:
        """
        execution_status = get_past_execution_status(config_unit, step_name)
        if execution_status is not None:
            return copy.deepcopy(execution_status)

        return self._get_cascaded_failure_execution_status(
            config_unit, step_name)

    def _get_cascaded_failure_execution_status(
            self,
            config_unit: ConfigUnit,
            step_name: str
            ) -> Optional[StepExecutionStatus]:
        blocking_execution_status = \
            self._get_blocking_failure_execution_status(config_unit, step_name)

        if blocking_execution_status is None:
            return None

        output = 'Previous failure of ' \
                 + blocking_execution_status.config_unit_name + ' / ' + \
                 blocking_execution_status.step_name

        if blocking_execution_status.restored:
            output += ' (restored)'

        return StepExecutionStatus(
            config_unit.name,
            step_name,
            ignored=True,
            successful=False,
            step_type=blocking_execution_status.step_type,
            output=output)

    def _get_blocking_failure_execution_status(
            self,
            config_unit: ConfigUnit,
            step_name: str
            ) -> Optional[StepExecutionStatus]:
        step_type = get_step_type(config_unit, step_name)

        if step_type == drkns.step.step_type.CLEANUP:
            return None

        if config_unit.name in self._failed_dependency_check_statuses:
            return self._failed_dependency_check_statuses[config_unit.name]

        if step_type == drkns.step.step_type.CHECK:
            return None

        # we are sure it is a "build" step now

        if config_unit.name in self._failed_dependency_build_statuses:
            return self._failed_dependency_build_statuses[config_unit.name]

        for dependency_unit in config_unit.dependencies:
            dependency_name = dependency_unit.name
            if dependency_name in self._failed_dependency_check_statuses:
                return self._failed_dependency_check_statuses[dependency_name]

        return None

    # END past status retrieval



import os
from subprocess import Popen
from typing import Optional, Union, Dict, List

from drkns.step.Step import Step
import drkns.step.step_type


class ConfigUnit:

    @staticmethod
    def _step_from_raw_steps(raw_steps: Dict[str, Union[str, Dict]]) \
            -> Dict[str, Step]:
        return {
            name: Step(command)
            for name, command in raw_steps.items()
        }

    def __init__(
            self,
            name: str,
            data: dict,
            ignored: List[str]
            ):
        self.name: str = name
        self.directory: str = data.get('directory', os.path.abspath('.'))

        self.check_steps: Dict[str, Step] = \
            self._step_from_raw_steps(data.get('checkSteps', {}))
        self.build_steps: Dict[str, Step] = \
            self._step_from_raw_steps(data.get('buildSteps', {}))
        self.cleanup_steps: Dict[str, Step] = \
            self._step_from_raw_steps(data.get('cleanupSteps', {}))

        self.dependencies: List[ConfigUnit] = data.get('dependencies', [])

        self.ignored: List[str] = ignored

        self.hash: Optional[str] = None

        self.pending_subprocesses: List[Popen] = []

    def __eq__(self, other):
        return self.name == other.name

    def get_dependency(self, dependency_name: str) -> Optional['ConfigUnit']:
        for dependency in self.dependencies:
            if dependency.name == dependency_name:
                return dependency

        return None

    def get_steps(self, step_type: str) -> Dict[str, Step]:
        drkns.step.step_type.check_step_type(step_type)

        if step_type == drkns.step.step_type.CHECK:
            return self.check_steps
        if step_type == drkns.step.step_type.BUILD:
            return self.build_steps
        if step_type == drkns.step.step_type.CLEANUP:
            return self.cleanup_steps

        raise Exception('Must have return something by now')



from typing import Optional

import os
from shutil import rmtree

from drkns.context.get_unit_step_path import \
    _get_path_to_unit, persistence_directory


def forget(unit_name: Optional[str] = None):
    if unit_name is None:
        rmtree(persistence_directory)

    unit_path = _get_path_to_unit(unit_name)  # type: ignore
    if not os.path.exists(unit_path):
        return

    rmtree(unit_path)



from typing import List, Tuple

from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.generation.GenerationTemplate import GenerationTemplate
from drkns.generation.generation.get_group_block import get_group_block
from drkns.generation.pattern_util \
    import format_list_in_template
from drkns.generation._tags import groups_tag, all_group_names_tag


def get_formatted_from_groups_configuration(
        generation_template: GenerationTemplate,
        groups: List[Tuple[str, List[ConfigUnit], List[str]]]
) -> str:
    group_blocks = []
    all_group_names = []
    for group_name, group_config_units, dependency_groups_name in groups:
        group_block = get_group_block(
            generation_template, group_name, group_config_units,
            dependency_groups_name)
        group_blocks.append(group_block)

        all_group_names.append(group_name)

    formatted = generation_template.template
    formatted = format_list_in_template(
        groups_tag, group_blocks, formatted, line_level=True)

    formatted = format_list_in_template(
        all_group_names_tag, all_group_names, formatted, line_level=False)

    return formatted



CHECK = 'check'
CLEANUP = 'cleanup'
BUILD = 'build'

types = [CHECK, CLEANUP, BUILD]


def check_step_type(step_type: str):
    if step_type in [CHECK, CLEANUP, BUILD]:
        return

    raise Exception('Unkown step_type: ' + step_type)



from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.generation.GenerationTemplate import GenerationTemplate
from drkns.generation.pattern_util import get_pattern_from_tag
from drkns.generation._tags import unit_name_tag


def get_unit_block(
        generation_template: GenerationTemplate,
        config_unit: ConfigUnit
) -> str:
    unit_name = config_unit.name
    unit_template = generation_template.unit_template
    unit_name_pattern = get_pattern_from_tag(unit_name_tag)
    return unit_template.replace(unit_name_pattern, unit_name)



import os
from typing import List, Optional

from drkns.configunit.ConfigUnit import ConfigUnit
import drkns.step.step_type


def get_error_string(config_unit: ConfigUnit, indent: Optional[int] = None) \
        -> Optional[str]:
    if indent is None:
        indent = 0

    errors = _get_errors(config_unit, indent + 1)

    if len(errors) == 0:
        return None

    original_indent_string = '   ' * indent
    indent_string = '   ' * (indent + 1)
    return original_indent_string + 'Error in "' + config_unit.name + '":\n' +\
        indent_string + ('\n' + indent_string).join(errors) + '\n\n'


def _get_errors(config_unit: ConfigUnit, indent: int) -> List[str]:
    errors: List[str] = []
    if config_unit.directory is None:
        errors.append(config_unit.name + ': directory is not set')
    elif not os.path.exists(config_unit.directory):
        errors.append(config_unit.name + ': directory does not exist')

    has_steps = (len(config_unit.check_steps) != 0) or \
                (len(config_unit.build_steps) != 0)
    has_dependencies = len(config_unit.dependencies) != 0
    if (not has_steps) and (not has_dependencies):
        message = config_unit.name + ': no steps nor dependencies are defined'
        errors.append(message)

    for dependency in config_unit.dependencies:
        dependency_error_string = get_error_string(dependency, indent)
        if dependency_error_string is not None:
            errors.append(dependency_error_string)

    names_error = _get_dependency_or_step_names_error(config_unit)
    if names_error is not None:
        errors.append(names_error)

    return errors


def _get_dependency_or_step_names_error(
        config_unit: ConfigUnit) -> Optional[str]:
    names = [dependency.name for dependency in config_unit.dependencies]
    for step_type in drkns.step.step_type.types:
        names += list(config_unit.get_steps(step_type).keys())

    duplicate_names = set()
    unique_names = set()
    for step_name in names:
        if step_name in duplicate_names:
            continue

        if step_name in unique_names:
            duplicate_names.add(step_name)
            continue

        unique_names.add(step_name)

    if len(duplicate_names) == 0:
        return None

    message = 'Config unit ' + config_unit.name + \
              ' reuses step or dependency names: ' \
              + ', '.join(duplicate_names)  # type: ignore
    return message



group_template_tag_prefix = 'DRKNS_GROUP_TEMPLATE_'  # BEGIN / END
unit_template_tag_prefix = 'DRKNS_UNIT_TEMPLATE_'  # BEGIN / END

groups_tag = 'DRKNS_GROUPS'
group_units_tag = 'DRKNS_GROUP_UNITS'

group_name_tag = 'DRKNS_GROUP_NAME'
unit_name_tag = 'DRKNS_UNIT_NAME'

dependency_groups_names_tag = 'DRKNS_DEPENDENCY_GROUPS_NAMES'
all_group_names_tag = 'DRKNS_ALL_GROUPS_NAMES'



from typing import Optional, List, Tuple

from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.runner.get_execution_plan import get_execution_plan
from drkns.runner.run_plan import run_plan
from drkns.runner.get_successful_flag_and_combined_output import\
    get_successful_flag_and_combined_output


def run(
        root_config_unit: ConfigUnit,
        target_step_name: Optional[str] = None,
        summary: bool = False,
        limit_output: bool = False) \
        -> Tuple[int, List[str]]:
    plan = get_execution_plan(root_config_unit, target_step_name)
    status_history = run_plan(plan)
    return get_successful_flag_and_combined_output(
        status_history, summary, limit_output)



import os
import pickle
from typing import Optional

from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.stepexecutionstatus.StepExecutionStatus import StepExecutionStatus
from drkns.context.Context import context
from drkns.context.get_unit_step_path import get_unit_step_path


def get_past_execution_status(
        config_unit: ConfigUnit,
        step_name: str) \
        -> Optional[StepExecutionStatus]:
    persistence_path = get_unit_step_path(config_unit, step_name)

    if persistence_path in context.cached_execution_status:
        return context.cached_execution_status[persistence_path]

    if os.path.exists(persistence_path):
        with open(persistence_path, 'rb') as persisted_file:
            step_execution_status = pickle.load(persisted_file)
            step_execution_status.restored = True
            context.cached_execution_status[persistence_path] =\
                step_execution_status
            return step_execution_status

    return None



from typing import Union, Dict


class Step:

    def __init__(self, data: Union[Dict,str]):
        if not isinstance(data, dict):
            data = {
                'command': data,
                'background': False
            }

        self.command: str = data['command']
        self.background: bool = data.get('background', False)



from typing import Tuple, List, Optional

from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.exception import CircularDependencyException, \
    UnknownDependencyException
import drkns.step.step_type
from drkns.step.get_step_type import get_step_type
from drkns.runner.remove_execution_plan_redundancies \
    import remove_execution_plan_redundancies


def get_execution_plan(
        config_unit: ConfigUnit,
        step_name: str = None) \
        -> List[Tuple[ConfigUnit, str, str]]:
    """
    Execution plan guarantees that steps are in dependency order:
    * Dependencies first
    * In order check steps, build steps and clean steps
    * If a specific steps is given, build steps not directly preceding it will
    be omitted

    :param config_unit:
    :param step_name:
    :return: config unit, step within the config unit, prefixed step name
    """
    # This public function only hide the parameters of the private one
    raw_execution_plan = _get_execution_plan(config_unit, step_name)
    execution_plan = remove_execution_plan_redundancies(raw_execution_plan)
    return execution_plan


def _get_execution_plan(
        config_unit: ConfigUnit,
        step_name: str = None,
        allowed_types: List[str] = None,
        prefix: str = None
        ) -> List[Tuple[ConfigUnit, str, str]]:
    if step_name is None:
        prefix = '' if prefix is None else prefix
        steps = _get_nested_dependencies_steps(config_unit, prefix)

        steps += _get_internal_steps(config_unit, prefix,
                                     allowed_types=allowed_types)
        return steps

    nested_config_unit, final_step_name = \
        _get_nested_config_unit_and_step(config_unit, step_name)
    prefix_parts = step_name.split('.')
    if final_step_name is not None:  # The final part is a step
        prefix_parts.pop()
    step_type = None
    if final_step_name is not None:
        step_type = get_step_type(nested_config_unit, final_step_name)

    step_prefix = '.'.join(prefix_parts)
    step_prefix += '.' if step_prefix else ''

    no_dependencies_types =\
        [drkns.step.step_type.CHECK, drkns.step.step_type.CLEANUP]
    if step_type in no_dependencies_types:
        return _get_internal_steps(
            nested_config_unit, step_prefix, final_step_name)

    # we know step_type is BUILD or None
    steps = _get_nested_dependencies_steps(nested_config_unit, step_prefix,
                                           [drkns.step.step_type.CHECK])
    steps += _get_internal_steps(nested_config_unit, step_prefix,
                                 target_step_name=step_name)
    return steps


def _get_nested_dependencies_steps(
        config_unit: ConfigUnit,
        prefix: str,
        allowed_types: List[str] = None
        ) -> List[Tuple[ConfigUnit, str, str]]:
    """
    If we are calling this function we want to resolve every dependency
    (at every steps i.e. check, build and cleanup)
    :param config_unit:
    :param prefix:
    :return:
    """
    # Circular dependency detection
    max_nesting = 10
    if prefix.count('.') > max_nesting:
        message = 'More than ' + str(max_nesting) +\
                  ' levels have been found, note that we do not support ' +\
                  'circular dependencies and believe that this is one instance.'
        raise CircularDependencyException(message)

    steps = []

    for dependency_unit in config_unit.dependencies:
        dependency_name = dependency_unit.name
        new_prefix = prefix + dependency_name + '.'
        steps += _get_execution_plan(
            dependency_unit, allowed_types=allowed_types, prefix=new_prefix)

    return steps


def _get_internal_steps(
        config_unit: ConfigUnit,
        prefix: str,
        target_step_name: str = None,
        allowed_types: List[str] = None
        ) -> List[Tuple[ConfigUnit, str, str]]:
    """

    :param config_unit:
    :param prefix:
    :param target_step_name: if specified will not include CHECK or BUILD steps
        after seeing this item
    :param allowed_types:
    :return:
    """
    step_names: List[str] = []

    allowed_types = drkns.step.step_type.types \
        if allowed_types is None else allowed_types

    has_check = drkns.step.step_type.CHECK in allowed_types
    has_build = drkns.step.step_type.BUILD in allowed_types

    if has_check or has_build:
        step_names += config_unit.get_steps(drkns.step.step_type.CHECK).keys()
    if has_build:
        step_names += config_unit.get_steps(drkns.step.step_type.BUILD).keys()

    # whatever happens we clean up
    step_names += config_unit.get_steps(drkns.step.step_type.CLEANUP).keys()

    plan_steps: List[Tuple[ConfigUnit, str, str]] = []
    found_target = False
    for step_name in step_names:
        is_cleanup = get_step_type(config_unit, step_name) == \
                     drkns.step.step_type.CLEANUP
        if found_target and not is_cleanup:
            continue

        prefixed_step_name = prefix if prefix is not None else ''
        prefixed_step_name += step_name
        plan_steps.append((config_unit, step_name, prefixed_step_name))
        if step_name == target_step_name:
            found_target = True

    return plan_steps


def _get_nested_config_unit_and_step(
        root_config_unit: ConfigUnit,
        composite_step_name: str) \
        -> Tuple[ConfigUnit, Optional[str]]:

    dot_index = composite_step_name.find('.')
    if dot_index == -1:
        dependency_config_unit = \
            root_config_unit.get_dependency(composite_step_name)
        if dependency_config_unit is not None:
            return dependency_config_unit, None
        return root_config_unit, composite_step_name

    dependency_name = composite_step_name[0:dot_index]
    dependency_config_unit = root_config_unit.get_dependency(dependency_name)
    if dependency_config_unit is None:
        message = dependency_name + ' is not a dependency in ' + \
                  root_config_unit.name
        raise UnknownDependencyException(message)

    sub_step_name = composite_step_name[dot_index + 1:]
    return _get_nested_config_unit_and_step(dependency_config_unit,
                                            sub_step_name)



from datetime import datetime
import os
from typing import Optional

import drkns.step.step_type

_cwd = os.getcwd()


class StepExecutionStatus:

    def __init__(
            self,
            config_unit_name: str,
            step_name: str,
            output: str,
            successful: bool = False,
            ignored: bool = False,
            step_type: str = drkns.step.step_type.CHECK
    ):
        self.config_unit_name: str = config_unit_name
        self.step_name: str = step_name
        self.successful: bool = successful
        self.ignored: bool = ignored
        self.step_type: str = step_type
        self.output: str = output

        self.restored: bool = False

        self.datetime: datetime = datetime.now()

    def name(self) -> str:
        return self.config_unit_name + '@' + self.step_name



from typing import List, Tuple

from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.runner.get_execution_plan import get_execution_plan
from drkns.configunit.get_hash import get_hash
from drkns.context.get_past_execution_status import get_past_execution_status


def get_debug_information(config_unit: ConfigUnit):
    execution_plan = get_execution_plan(config_unit)
    output_elements = [
        _get_hash_debug_information(execution_plan),
        _get_execution_plan_debug_information(execution_plan)
    ]

    return '\n\n'.join(output_elements)


def _get_hash_debug_information(
        execution_plan: List[Tuple[ConfigUnit, str, str]]
        ) -> str:
    names_to_hashes = {}

    for (config_unit, _, _) in execution_plan:
        if config_unit.name in names_to_hashes:
            continue

        names_to_hashes[config_unit.name] = get_hash(config_unit)

    lines = ['* ' + name + ': \n\t' + unit_hash
             for name, unit_hash in names_to_hashes.items()]
    lines.insert(0, 'Config hashes:')
    return '\n'.join(lines)


def _get_execution_plan_debug_information(
        execution_plan: List[Tuple[ConfigUnit, str, str]]
        ) -> str:
    lines = []

    for (config_unit, step_name, prefixed_step_name) in execution_plan:
        status = get_past_execution_status(config_unit, step_name)
        line = '* ' + prefixed_step_name + ': \n\t'

        if status is not None:
            line += 'previous execution as '
            if status.successful:
                line += 'successful'
            else:
                if status.ignored:
                    line += 'ignored'
                else:
                    line += 'execution error'
        else:
            line += '---'

        lines.append(line)

    lines.insert(0, 'Execution plan:')
    return '\n'.join(lines)


from typing import Tuple


from drkns.context.get_unit_step_path import persistence_directory
from drkns.util import sh


def sync_in(target_s3_path: str) -> Tuple[int, str]:
    command = 'aws s3 sync {} {}'.format(target_s3_path, persistence_directory)
    status, output, _ = sh(command)
    return status, output


def sync_out(target_s3_path: str, delete: bool) -> Tuple[int, str]:
    delete_flag = ''
    if delete:
        delete_flag = '--delete'

    command = 'aws s3 sync {} {} {}'.format(
        persistence_directory, target_s3_path, delete_flag)

    status, output, _ = sh(command)
    return status, output



from typing import List

import os


from drkns.util import get_longest_common_prefix
from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.generation.GenerationTemplate import GenerationTemplate
from drkns.generation.generation.get_unit_block import get_unit_block
from drkns.generation.pattern_util \
    import format_list_in_template, get_pattern_from_tag, \
        format_list_optional_in_template
from drkns.generation._tags \
    import group_units_tag, group_name_tag, dependency_groups_names_tag


def get_group_block(
        generation_template: GenerationTemplate,
        group_name: str,
        group_config_units: List[ConfigUnit],
        dependency_group_names: List[str]
) -> str:
    formatted_config_units = [
        get_unit_block(generation_template, config_unit)
        for config_unit in group_config_units
    ]
    formatted_group = generation_template.group_template
    formatted_group = format_list_in_template(
        group_units_tag, formatted_config_units, formatted_group,
        line_level=True)
    formatted_group = format_list_in_template(
        dependency_groups_names_tag,
        dependency_group_names,
        formatted_group, line_level=False)
    formatted_group = format_list_optional_in_template(
        dependency_groups_names_tag,dependency_group_names, formatted_group)

    formatted_group = formatted_group.replace(
        get_pattern_from_tag(group_name_tag), group_name)

    return formatted_group


def _get_group_name(
        generation_template: GenerationTemplate,
        group_config_units: List[ConfigUnit]
) -> str:
    truncated_directories = []
    names = []

    template_root_directory = os.path.abspath(generation_template.source_path)
    for config_unit in group_config_units:
        unit_directory = os.path.abspath(config_unit.directory)
        unit_directory = unit_directory.replace(
            template_root_directory, unit_directory)
        truncated_directories.append(unit_directory)

        names.append(config_unit.name)

    prefix = get_longest_common_prefix(truncated_directories)
    names_string = ', '.join(names)
    if len(names) > 1:
        names_string = '{' + names_string + '}'

    return prefix + names_string








import os

from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.configunit.get_hash import get_hash
from drkns.step.get_step_type import get_step_type

persistence_directory = '.drknspersistence'
extension = '.drknsdata'


def get_unit_step_path(config_unit: ConfigUnit, step_name: str):
    prefix = get_step_type(config_unit, step_name) + '_'

    unit_path = _get_path_to_unit(config_unit.name)
    unit_directory = os.path.join(unit_path, prefix + step_name)
    if not os.path.exists(unit_directory):
        os.makedirs(unit_directory, exist_ok=True)

    return os.path.join(unit_directory, get_hash(config_unit) + extension)


def _get_path_to_unit(unit_name: str) -> str:
    unit_name = unit_name.replace('/', '_')
    unit_name = unit_name.replace('..', '_')
    return os.path.join(persistence_directory, unit_name)



from dirhash import dirhash  # type: ignore
from hashlib import sha1

from drkns.configunit.ConfigUnit import ConfigUnit


def get_hash(config_unit: ConfigUnit) -> str:
    if config_unit.hash is None:
        config_unit.hash = _get_computed_hash(config_unit)

    return config_unit.hash


def _get_computed_hash(config_unit: ConfigUnit) -> str:
    hash_input = \
        dirhash(config_unit.directory, 'sha1',
                ignore=config_unit.ignored)

    for dependency_config_unit in config_unit.dependencies:
        hash_input += _get_computed_hash(dependency_config_unit)

    hashed = hash_input[0:7] + '-' + \
        sha1(hash_input.encode('utf-8')).hexdigest()
    return hashed



from typing import Tuple, List

from drkns.configunit.ConfigUnit import ConfigUnit


def remove_execution_plan_redundancies(
        execution_plan: List[Tuple[ConfigUnit, str, str]]
        ) -> List[Tuple[ConfigUnit, str, str]]:
    cleaned_plan = []
    found_steps = {}
    for execution_step in execution_plan:
        step_key = _get_execution_step_key(execution_step)
        if step_key in found_steps:
            continue

        cleaned_plan.append(execution_step)
        found_steps[step_key] = True

    return cleaned_plan


def _get_execution_step_key(
        execution_step: Tuple[ConfigUnit, str, str]
        ) -> str:
    config_unit, step_name, _ = execution_step
    return config_unit.name + '.' + step_name + '@' + config_unit.directory



import sys
from typing import Collection
import os

from drkns.exception import MissingCommandException, UnknownCommandException,\
    MissingSyncDirectionException, MissingS3PathException,\
    UnknownCommandFlagException, MissingForgetTargetException
from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.configunit.load import load
from drkns.configunit.get_error_string import get_error_string
from drkns.context.clean_persistence_files import clean_persistence_files
from drkns.context.sync import sync_in, sync_out
from drkns.context.forget import forget
from drkns.runner.run import run
from drkns.runner.get_execution_plan import get_execution_plan
from drkns.runner.get_dependency_groups_configuration \
    import get_dependency_groups_configuration
from drkns.generation.templateloading.get_generation_template \
    import get_generation_template
from drkns.generation.generation.get_formatted_from_groups_configuration \
    import get_formatted_from_groups_configuration
from drkns.debug.get_debug_information import get_debug_information


class Cli:

    def __init__(self):
        self._args: Collection[str] = []
        self._command: str = None
        self._force_success: bool = False
        self._summary: bool = False
        self._limit_output: bool = False
        self._delete: bool = False

    def handle(self):
        self._parse_args()

        check_command = 'check'
        if self._command == check_command:
            self._check()
            return

        clean_command = 'clean'
        if self._command == clean_command:
            self._clean()
            return

        debug_command = 'debug'
        if self._command == debug_command:
            self._debug()
            return

        forget_command = 'forget'
        if self._command == forget_command:
            self._forget()
            return

        list_command = 'list'
        if self._command == list_command:
            self._list()
            return

        run_command = 'run'
        if self._command == run_command:
            self._run()
            return

        sync_command = 'sync'
        if self._command == sync_command:
            self._sync()
            return

        generate_command = 'generate'
        if self._command == generate_command:
            self._generate()
            return

        allowed_commands = [
            check_command, clean_command, debug_command, forget_command,
            list_command, run_command, sync_command, generate_command
        ]

        message = 'Unkown command: ' + self._command +\
                  '. Allowed commands are "' + '", "'.join(allowed_commands) + \
                  '".'
        raise UnknownCommandException(message)

    def _parse_args(self):
        provided_args = sys.argv[1:]
        if len(provided_args) == 0:
            raise MissingCommandException()

        self._command = provided_args.pop(0)

        force_success_flag = '--force-success'
        if force_success_flag in provided_args:
            index = provided_args.index(force_success_flag)
            provided_args.pop(index)
            self._force_success = True

        summary_flag = '--summary'
        if summary_flag in provided_args:
            index = provided_args.index(summary_flag)
            provided_args.pop(index)
            self._summary = True

        limit_output_flag = '--limit-output'
        if limit_output_flag in provided_args:
            index = provided_args.index(limit_output_flag)
            provided_args.pop(index)
            self._limit_output = True

        delete_flag = '--delete'
        if delete_flag in provided_args:
            index = provided_args.index(delete_flag)
            provided_args.pop(index)
            self._delete = False

        for arg in provided_args:
            if arg.find('--') != -1:
                allowed_flags = [
                    force_success_flag, summary_flag, limit_output_flag,
                    delete_flag
                ]

                message = 'Unkown command flag: ' + arg + \
                          '. Allowed flags are "' + \
                          '", "'.join(allowed_flags) + '".'
                raise UnknownCommandFlagException(message)

        self._args = provided_args

    def _check(self):
        config_unit = self._get_config_unit()
        error_string = get_error_string(config_unit)

        if error_string is None:
            sys.exit()

        sys.exit(error_string)

    # noinspection PyMethodMayBeStatic
    def _clean(self):
        clean_persistence_files()
        sys.exit()

    def _debug(self):
        config_unit = self._get_config_unit()
        print(get_debug_information(config_unit))
        sys.exit()

    def _forget(self):
        if len(self._args) != 1:
            error_message = 'Missing forget target. Allowed targets are a ' +\
                'unit name, "main" and "all".'
            raise MissingForgetTargetException(error_message)

        unit_name = self._args[0]
        if unit_name == 'all':
            unit_name = None

        forget(unit_name)

    def _list(self):
        config_unit = self._get_config_unit()
        execution_plan = get_execution_plan(config_unit)
        step_names = \
            [prefixed_step_name for _, _, prefixed_step_name in execution_plan]
        print('\n'.join(step_names))

    def _run(self):
        config_unit = self._get_config_unit()

        target = None
        if len(self._args) > 0:
            target = self._args[0]

        successful, output_lines = run(config_unit, target,
                                       self._summary, self._limit_output)
        if len(output_lines) > 0:
            print('\n'.join(output_lines))
        if successful or self._force_success:
            sys.exit()

        sys.exit(1)

    def _sync(self):
        if (len(self._args) < 1) \
                or (self._args[0] not in ['in', 'out']):
            message = 'Sync direction should be in or out'
            raise MissingSyncDirectionException(message)

        sync_direction = self._args[0]

        environment_variable_name = 'DRKNS_S3_PATH'
        target_s3_path = os.environ.get(environment_variable_name)
        if len(self._args) == 2:
            target_s3_path = self._args[1]

        if target_s3_path is None:
            message = 'Missing s3 path, it can be provided as an argument or ' \
                      'an environment variable: ' + environment_variable_name
            raise MissingS3PathException(message)

        if sync_direction == 'in':
            (return_code, output) = sync_in(target_s3_path)
        else:
            (return_code, output) = sync_out(target_s3_path, self._delete)

        successful = return_code == 0

        if not successful:
            print(output)

        sys.exit(return_code)

    def _generate(self):
        config_unit = self._get_config_unit()
        generation_template = get_generation_template('.')
        groups_configuration = get_dependency_groups_configuration(config_unit)
        generated = get_formatted_from_groups_configuration(
            generation_template,
            groups_configuration)
        print(generated)

    # noinspection PyMethodMayBeStatic
    def _get_config_unit(self) -> ConfigUnit:
        return load('drkns.yml')



from subprocess import Popen, PIPE, STDOUT
import sys
from typing import Tuple, Optional, List


class BColors:

    # source: https://stackoverflow.com/questions/287871
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def sh(command: str, detached: Optional[bool] = None)\
        -> Tuple[int, str, Popen]:
    kwargs = {
        'shell': True,
        'stdin': PIPE,
        'stdout': PIPE,
        'stderr': STDOUT
    }

    p = Popen(command, **kwargs)  # type: ignore

    if detached:
        return 0, 'Detached process', p

    p_stdout = p.communicate()[0]
    if p_stdout is not None:
        p_stdout = p_stdout.decode(sys.getdefaultencoding(), 'ignore')

    return p.returncode, p_stdout, p


def get_longest_common_prefix(items: List[str]) -> str:
    longest_prefix = ''
    if len(items) == 0:
        return longest_prefix

    max_index = min([len(item) for item in items])
    for i in range(0, max_index):
        reference_char = items[0][i]
        for item in items:
            if item[i] != reference_char:
                return longest_prefix

        longest_prefix += reference_char

    return longest_prefix



from os import walk, path, unlink
import pickle
import datetime

from drkns.context.get_unit_step_path import persistence_directory, extension
from drkns.stepexecutionstatus.StepExecutionStatus import StepExecutionStatus


def clean_persistence_files():
    if not path.exists(persistence_directory):
        return

    too_old_file_paths = _get_too_old_file_paths()

    for too_old_file_path in too_old_file_paths:
        unlink(too_old_file_path)


def _get_too_old_file_paths():
    if not path.exists(persistence_directory):
        return

    too_old_paths = []

    for (dir_path, _, filenames) in walk(persistence_directory):
        if len(filenames) < 3:  # This step is not executed very often
            continue

        for filename in filenames:
            if path.splitext(filename)[-1] != extension:
                continue

            file_path = path.join(dir_path, filename)

            if _is_file_to_old(file_path):
                too_old_paths.append(file_path)

    return too_old_paths


def _is_file_to_old(file_path) -> bool:
    status = _get_persisted_status(file_path)
    a_week_ago = datetime.datetime.now() - datetime.timedelta(weeks=1)
    return status.datetime < a_week_ago


def _get_persisted_status(file_path) -> StepExecutionStatus:
    with open(file_path, 'rb') as persisted_file:
        status = pickle.load(persisted_file)
        return status



from distutils.core import setup
from setuptools import find_packages

with open('../README.md') as readme_file:
    long_description = readme_file.read()

setup(
    name='drkns',
    version='3.0.3',
    description='Simple monorepo build tool',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/frantzmiccoli/drkns',
    packages=find_packages('./'),
    package_dir={'./src': 'src'},
    license='MIT',
    scripts=['../bin/drkns'],
    install_requires=[
        'awscli>=1,<2',
        'dirhash>=0.2,<1.0',
        'PyYAML>=5.3,<6.0'  #,
        #'boto3>=1.12.0,<2.0.0'
    ],
    keywords="devops build test monorepo cache",
)



from typing import List, Tuple


from drkns.configunit.ConfigUnit import ConfigUnit
from drkns.exception import DependenciesNotAvailable
from drkns.runner.get_execution_plan import get_execution_plan


class _DependencyGroupsConfigurationResolver:

    def __init__(self, root_config_unit: ConfigUnit):
        self._root_config_unit: ConfigUnit = root_config_unit

        self._execution_plan_units: List[ConfigUnit] = []
        self._groups_configuration: \
            List[Tuple[str, List[ConfigUnit], List[str]]] = []
        self._available_dependencies: List[ConfigUnit] = []

    def get_groups_configuration(
            self
    ) -> List[Tuple[str, List[ConfigUnit], List[str]]]:
        self._load_execution_plan_units()
        self._resolve_group_configurations()

        return self._groups_configuration


    def _load_execution_plan_units(self):
        execution_plan = get_execution_plan(self._root_config_unit)

        for config_unit, _, _ in execution_plan:
            if config_unit in self._execution_plan_units:
                continue

            self._execution_plan_units.append(config_unit)

    def _resolve_group_configurations(self):
        while len(self._execution_plan_units) > 0:
            config_unit = self._execution_plan_units.pop(0)
            if not self._are_dependencies_available(config_unit):
                error_message = \
                    'This should never occur as we are proceeding through ' + \
                    'the ordered execution plan'

                raise DependenciesNotAvailable(error_message)

            group_name = config_unit.name
            group_units = [config_unit]
            self._available_dependencies.append(config_unit)
            dependency_groups_names = \
                [dependency.name
                 for dependency in config_unit.dependencies]
            group_configuration = \
                (group_name, group_units, dependency_groups_names)
            self._groups_configuration.append(
                group_configuration)

    def _are_dependencies_available(self, config_unit: ConfigUnit):
        for dependency in config_unit.dependencies:
            if dependency not in self._available_dependencies:
                return False

        return True


def get_dependency_groups_configuration(
        root_config: ConfigUnit
) -> List[Tuple[str, List[ConfigUnit], List[str]]]:
    resolver = _DependencyGroupsConfigurationResolver(root_config)
    return resolver.get_groups_configuration()



import re
from typing import Optional


from drkns.exception import MissingGenerationTemplateTag
from drkns.generation.GenerationTemplate import GenerationTemplate
import drkns.generation.pattern_util as pattern_util
from drkns.generation._tags \
    import group_units_tag, group_name_tag, unit_name_tag, \
    dependency_groups_names_tag, all_group_names_tag


def check_generation_template(generation_template: GenerationTemplate):
    errors_candidates = []

    group_name_tag_error = _get_tag_error(
        group_name_tag, generation_template, 'group_template', 'static')
    errors_candidates.append(group_name_tag_error)

    group_units_tag_error = _get_tag_error(
        group_units_tag, generation_template, 'group_template', 'list')
    errors_candidates.append(group_units_tag_error)

    unit_name_tag_error = _get_tag_error(
        unit_name_tag, generation_template, 'unit_template', 'static')
    errors_candidates.append(unit_name_tag_error)

    dependency_groups_names_tag_error = _get_tag_error(
        dependency_groups_names_tag, generation_template, 'group_template',
        'list')
    errors_candidates.append(dependency_groups_names_tag_error)

    all_group_names_tag_error = \
        _get_tag_error(all_group_names_tag, generation_template,
                       'template', 'list')
    errors_candidates.append(all_group_names_tag_error)

    errors = [error for error in errors_candidates if error is not None]
    if len(errors) > 0:
        raise MissingGenerationTemplateTag('\n'.join(errors))


def _get_tag_error(
        tag: str,
        generation_template: GenerationTemplate,
        field: str,
        tag_type: str
) -> Optional[str]:
    search_in_template = getattr(generation_template, field)
    if tag_type == 'static':
        pattern = pattern_util.get_pattern_from_tag(tag)
        if pattern in search_in_template:
            return None
        return 'Missing ' + pattern + ' in ' + field

    tag_re = pattern_util.get_list_re_from_tag(tag, False)
    match = tag_re.search(search_in_template)
    if match is not None:
        return None
    return 'Missing ' + tag_re.pattern + ' in ' + field

